import 'dart:math';
import 'dart:typed_data';

import 'package:bitcoin_flutter/bitcoin_flutter.dart' as bitcoin_flutter;
import 'package:bip39/bip39.dart' as bip39;
import 'package:bs58check/bs58check.dart';
import 'package:walletpackage/src/bitcoinTransaction.dart';
import 'package:walletpackage/src/ethereumTransaction.dart';
import 'package:walletpackage/src/tronTransactioin.dart';

import 'package:crypto/crypto.dart';
import 'package:web3dart/crypto.dart' as web3dart_crypto;
import 'package:web3dart/web3dart.dart' as web3dart;

class WalletData {
  String mnemonic; //助记词，本地不保存
  String keystore;
  Map<String, String> addresses = new Map<String, String>();
}

class CoinInfo {
  final String path;
  final Object network; //NetworkType or int
  final int decimals;
  final String address;
  final int gasLimit;

  CoinInfo(this.path, this.network, this.decimals,
      {this.address, this.gasLimit});
}

class WalletProxy {
  static final _litcoin = new bitcoin_flutter.NetworkType(
      messagePrefix: '\x19Litecoin Signed Message:\n',
      bech32: 'ltc',
      bip32: new bitcoin_flutter.Bip32Type(
          public: 0x019da462, private: 0x019d9cfe),
      pubKeyHash: 0x30,
      scriptHash: 0x32,
      wif: 0xb0);

  static final _bitcoincash = bitcoin_flutter.NetworkType(
      messagePrefix: '\x18Bitcoin Signed Message:\n',
      bech32: 'bc',
      bip32: new bitcoin_flutter.Bip32Type(
          public: 0x0488b21e, private: 0x0488ade4),
      pubKeyHash: 0x00,
      scriptHash: 0x05,
      wif: 0x80);

  static final Map<String, CoinInfo> bitcoinType = {
    'BTC': CoinInfo("m/44'/0'/0'/0/0", bitcoin_flutter.bitcoin, 8),
    'LTC': CoinInfo("m/44'/2'/0'/0/0", _litcoin, 8),
    'BCH': CoinInfo("m/44'/145'/0'/0/0", _bitcoincash, 8),
  };
  static final Map<String, CoinInfo> omniType = {
    'USDT':
        CoinInfo("m/44'/0'/0'/0/0", bitcoin_flutter.bitcoin, 8, address: '31'),
    'TUSDT':
        CoinInfo("m/44'/0'/0'/0/1", bitcoin_flutter.testnet, 8, address: '2'),
  };
  static final Map<String, CoinInfo> ethereumType = {
    'ETH': CoinInfo("m/44'/60'/0'/0/0", 1, 18, gasLimit: 21000),
    'ETC': CoinInfo("m/44'/61'/0'/0/0", 1, 18, gasLimit: 21000),
  };
  static final Map<String, CoinInfo> erc20Type = {
    'EUSDT': CoinInfo("m/44'/60'/0'/0/0", 1, 6,
        address: '0xdAC17F958D2ee523a2206206994597C13D831ec7', gasLimit: 60000),
    'SIM': CoinInfo("m/44'/60'/0'/0/1", 4, 18,
        address: '0x454eb8D2994730a840A37e2A937e525916842Db8', gasLimit: 60000)
  };
  static final Map<String, CoinInfo> tronType = {
    'TRX': CoinInfo("m/44'/195'/0'/0/0", 1, 6, gasLimit: 2000000)
  };
  static final Map<String, CoinInfo> trc20Type = {
    'TRCUSDT': CoinInfo("m/44'/195'/0'/0/0", 1, 6,
        address: 'TR7NHqjeKQxGTCi8q8ZY4pL8otSzgjLj6t',
        gasLimit: 2000000),
    'TRCTUSDT': CoinInfo("m/44'/195'/0'/0/0", 1, 6,
        address: 'TRT9iiqp6GQWxXK65GYjx3QyvvJreyatSq',
        gasLimit: 2000000)
  };
  static final Map<String, CoinInfo> _allCoinType = new Map<String, CoinInfo>();

  static bool checkType(String type) {
    if (_allCoinType.length == 0) {
      _allCoinType
        ..addAll(bitcoinType)
        ..addAll(ethereumType)
        ..addAll(omniType)
        ..addAll(erc20Type)
        ..addAll(tronType)
        ..addAll(trc20Type);
    }
    return _allCoinType.containsKey(type);
  }

  static bitcoin_flutter.HDWallet _getChildNode(
      bitcoin_flutter.HDWallet hdWallet, String type) {
    if (checkType(type)) {
      CoinInfo coinInfo = _allCoinType[type];
      // print(coinInfo.path);
      hdWallet.network = (coinInfo.network is bitcoin_flutter.NetworkType
          ? coinInfo.network
          : null);
      bitcoin_flutter.HDWallet childnode = hdWallet.derivePath(coinInfo.path);
      return childnode;
    } else {
      print('WalletProxy _getChildNode _allType not contains key : $type');
      return null;
    }
  }

  static String _node2EtherumAddress(bitcoin_flutter.HDWallet ethnode) {
    if (ethnode == null) return null;
    web3dart.EthereumAddress ethaddr = web3dart.EthereumAddress(web3dart_crypto
        .publicKeyToAddress(web3dart_crypto.privateKeyBytesToPublic(
            web3dart_crypto.hexToBytes(ethnode.privKey))));
    return ethaddr.hexEip55;
  }

  static String _node2TronAddress(bitcoin_flutter.HDWallet tronnode) {
    if (tronnode == null) return null;
    // print(tronnode.privKey);
    Uint8List pubKey = web3dart_crypto
        .privateKeyBytesToPublic(web3dart_crypto.hexToBytes(tronnode.privKey));
    String pubKeyKeccak256 =
        web3dart_crypto.bytesToHex(web3dart_crypto.keccak256(pubKey));
    String addressHex = '41' + pubKeyKeccak256.substring(24);
    // print(addressHex);
    Uint8List addressBin = web3dart_crypto.hexToBytes(addressHex);
    String hash1 = sha256.convert(sha256.convert(addressBin).bytes).toString();
    String checksum = hash1.substring(0, 8);
    checksum = addressHex + checksum;
    String addr = base58.encode(web3dart_crypto.hexToBytes(checksum));
    // print(addr);
    return addr;
  }

  static void _addBitcoinAddress(
      WalletData wd, bitcoin_flutter.HDWallet hdWallet, String type) {
    bitcoin_flutter.HDWallet node = _getChildNode(hdWallet, type);
    wd.addresses.addAll({type: node.address});
  }

  static void _addEtherumAddress(
      WalletData wd, bitcoin_flutter.HDWallet hdWallet, String type) {
    bitcoin_flutter.HDWallet node = _getChildNode(hdWallet, type);
    wd.addresses.addAll({type: _node2EtherumAddress(node)});
  }

  static void _addTronAddress(
      WalletData wd, bitcoin_flutter.HDWallet hdWallet, String type) {
    bitcoin_flutter.HDWallet node = _getChildNode(hdWallet, type);
    wd.addresses.addAll({type: _node2TronAddress(node)});
  }

  static Future<WalletData> createMnemonicAndAddress(String passwd) async {
    String mnemonic = bip39.generateMnemonic();
    WalletData walletData =
        await importMnemonicAndAddress(mnemonic, passwd: passwd);
    return walletData;
  }

  static Future<WalletData> importMnemonicAndAddress(String mnemonic,
      {String passwd}) async {
    if (bip39.validateMnemonic(mnemonic)) {
      if (passwd == null) {
        passwd = Random.secure().nextInt(999999).toString();
        print(passwd);
      }
      WalletData wd = new WalletData();
      wd.mnemonic = mnemonic;

      bitcoin_flutter.HDWallet hdWallet =
          bitcoin_flutter.HDWallet.fromSeed(bip39.mnemonicToSeed(wd.mnemonic));

      web3dart.Wallet wallet = web3dart.Wallet.createNew(
          web3dart.EthPrivateKey.fromHex(hdWallet.seed),
          passwd,
          Random.secure());
      wd.keystore = wallet.toJson();
//      print(wd.keystore);

      bitcoinType.keys.forEach((type) {
        _addBitcoinAddress(wd, hdWallet, type);
      });
      omniType.keys.forEach((type) {
        _addBitcoinAddress(wd, hdWallet, type);
      });
      ethereumType.keys.forEach((type) {
        _addEtherumAddress(wd, hdWallet, type);
      });
      erc20Type.keys.forEach((type) {
        _addEtherumAddress(wd, hdWallet, type);
      });
      tronType.keys.forEach((type) {
        _addTronAddress(wd, hdWallet, type);
      });
      trc20Type.keys.forEach((type) {
        _addTronAddress(wd, hdWallet, type);
      });
//      print(wd.addresses);

      return wd;
    } else {
      return null;
    }
  }

  static Future<String> changePassword(
      String keystore, String passwd, String newPasswd) async {
    try {
      web3dart.Wallet wallet = web3dart.Wallet.fromJson(keystore, passwd);
      web3dart.Wallet newWallet = web3dart.Wallet.createNew(
          wallet.privateKey, newPasswd, Random.secure());
      return newWallet.toJson();
    } catch (e) {
      print('WalletProxy.changePassword error : $e');
    }
    return null;
  }

  static Future<String> signTransaction(
      String type, String keystore, String passwd, Object tx) async {
    try {
      web3dart.Wallet wallet = web3dart.Wallet.fromJson(keystore, passwd);
      bitcoin_flutter.HDWallet hdWallet;
      bitcoin_flutter.HDWallet childWallet;
      String signedTxhex;
      if (checkType(type)) {
        if (tx is Btransaction) {
          hdWallet = bitcoin_flutter.HDWallet.fromSeed(
              wallet.privateKey.privateKey,
              network: tx.network);
          childWallet = _getChildNode(hdWallet, type);
          print(childWallet.address);
          signedTxhex = await BitcoinTransaction.signBitcoinTransaction(
              childWallet.wif, tx);
        } else if (tx is Etransaction) {
          hdWallet =
              bitcoin_flutter.HDWallet.fromSeed(wallet.privateKey.privateKey);
          childWallet = _getChildNode(hdWallet, type);
          print(_node2EtherumAddress(childWallet));
          signedTxhex = await EthereumTransaction.signEthereumTransaction(
              childWallet.privKey, tx);
        } else if (tx is Ttransaction) {
          hdWallet =
              bitcoin_flutter.HDWallet.fromSeed(wallet.privateKey.privateKey);
          childWallet = _getChildNode(hdWallet, type);
          print(_node2TronAddress(childWallet));
          signedTxhex = await TronTransaction.signTronTransaction(
              childWallet.privKey, tx);
        } else {
          print('WalletProxy.signTransaction unknown tx type : $tx');
        }
      } else {
        print('WalletProxy.signTransaction unknown coin type : $type');
      }
      return signedTxhex;
    } catch (e) {
      print('WalletProxy.signTransaction error : $e');
    }
    return null;
  }
}
